/*
 ============================================================================
 mzscript - an advanced script module for LiteStep/PureLS
 ============================================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 ============================================================================

 mzscript 0.9 jam 3 (2002-06-14) / Phil
    - Fixed message boxes appearing beneath all other windows.
 mzscript 0.9 jam 2 (2002-06-14) / Tres`ni
    - Fixed null string treated as zero bug (isnum)
 mzscript 0.9 beta 9 aka jam (2002-06-10) / Tres`ni
    - Added modulas operationgs (!varMod)
	- Added random value (!varRnd)
	- Added %{milli} - Returns Milliseconds
 mzscript 0.9 preview (2002-04-12) / ilmcuts
	- many internal changes (e.g. new, faster parser) [thanks Smoogle]
	- vars are now defined in separate files (*mzVarFile)
	- multiple script files allowed (*mzScriptFile)
	- vars added at runtime cannot be saved (mzAutosaveVars & !varSaveAll syntax changed)
	- script *start items are now executed last
		- mzNoStartOnRefresh will disable the *start items on !refresh
	- %{} is now the default syntax
		- use mzScriptOldSyntax to work with %[].
	- misc minor changes
	- disabled in the preview: !varSave, !varReplace, !scriptload
	- readme still incomplete but should cover major changes

 mzscript 0.8.7 (2002-04-04) / ilmcuts (unreleased)
	- added %[bpp] - this returns the current bpp (32 bit, 16 bit...)
	
 mzscript 0.8.6l (2002-03-12) / Phil@leaf
    - New syntax to workaround possible conflicts with ckDialog, etc.
    - Added support for new syntax/old syntax.
        - Defaults to new syntax support.

 mzscript 0.8.5l (2002-03-07) / Phil@leaf
    - Added a function to turn variables from floats to integers.
	- Call !VarInt with the name of the variable.
    - Added code to correctly round up/down the integers based on float values.

 mzscript 0.8l (2001-10-23) / Phil@leaf
    - Project settings changed to produce a multithreaded DLL.
	  - Size of compiled DLL has been halved with this change.
	- Compiled against most recent CVS of LiteStep 0.24.6.

 mzscript 0.8 (2001-09-24) / qwilk
	- Now works with PureLS! (thanks to jugg)
	- Added !Refresh support
	- Added !Pause (courtesy of Maduin's pause.dll)
	- Made changes to make mzscript compile with current LSAPI
	- Cleaned up the source code

 mzscript 0.7.2 (2000-05-22) / Maze
	- First open source release (see readme for previous history)

 ============================================================================ */
 
#include "mzscript.h"

#pragma warning ( disable : 4100 )


//===========================================================================

char* szAppName = "mzscript";
char* szVersion = "mzscript.dll: 0.9 beta 11 (maze/qwilk/leaf/ilmcuts/Tres`ni/phil)";
char* szDetailed = "mzscript.dll version 0.9 beta 11 date 2003-03-13 (maze/qwilk/leaf/ilmcuts/Tres`ni/phil)";
char internalname[] = "mzinternalvar"; // filename for internal vars that won't be saved to any file

bangcmddef Bangs[] = {
	{"!varSet",		bangSetVar},
    {"!varRemove",	bangRemoveVar},
    {"!varShow",	bangShowVar},
    {"!varRun",		bangRunVar},
    {"!varAdd",		bangAddVar},
    {"!varMul",		bangMulVar},
    {"!varInt",		bangIntVar},
	{"!varMod",		bangModVar},
	{"!varRnd",		bangRndVar},

    {"!varDump",	bangVarDump},

    {"!varSave",	bangSaveVar},
//    {"!varReplace",	bangReplaceVar},
    {"!varSaveAll",	bangSaveAllVars},

    {"!setlistsep",	bangSetListSep},

    {"!ifExist",	bangIfExist},
    {"!ifNExist",	bangIfNExist},
    {"!ifEq",		bangIfEq},
    {"!ifNEq",		bangIfNEq},
    {"!ifEval",		bangIfEval},

    {"!exec",		bangExec},
    {"!msgbox",		bangMsgBox},

    {"!scriptload",	readScriptFile},
    {"!scriptremove", bangRemoveScript},
	{"!pause", pause},
	{NULL, NULL}
};

//===========================================================================
// Module initialization and cleanup
//===========================================================================

int initModuleEx(HWND parent, HINSTANCE dllInst, LPCSTR szPath)
{
	WNDCLASS wc;
	hwndParent = parent;
	hInstance = dllInst;
	memset(&wc,0,sizeof(wc));
	wc.lpfnWndProc = WndProc;       // our window procedure
	wc.hInstance = dllInst;         // hInstance of DLL
	wc.lpszClassName = szAppName;   // our window class name
	
	if (!RegisterClass(&wc)) 
	{
		errorMsg("Error: Could not register window class");
		return 1;
	}

	hwndMain = CreateWindowEx(WS_EX_TOOLWINDOW, szAppName, szAppName, WS_CHILD, 0, 0, 0, 0, parent, NULL, dllInst, NULL);
	if (!hwndMain) return 1;
	SetWindowLong(hwndMain, GWL_USERDATA, magicDWord);
	SendMessage(hwndParent, LM_REGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);

	// code taken from RabidModuleBangs
	bangcmddef *pBang;

	pBang = Bangs;
	while (pBang->Name != NULL) {
		AddBangCommand(pBang->Name, pBang->Command);
		pBang++;
	}

	/*If you add time.h delete these 3 lines */
	SYSTEMTIME systemTime;
	GetLocalTime(&systemTime);
	/* Gotta love this math, figures an estimated seconds that have elapsed since 0AD or so */
	srand(((((((((((systemTime.wYear * 12) + systemTime.wMonth) * 30) + systemTime.wDay) * 24) + systemTime.wHour) * 60) + systemTime.wMinute) * 60) + systemTime.wSecond));

	/* Use with time.h and delete all the math above
	srand(time(0));
	*/

//    return LoadSetup();
	int code;
	code = LoadSetup();
	if (bReadStep) readStep(parent);
	return code;
}

//--------------------

void quitModule(HINSTANCE dll)
{
	if (autosave==1) bangSaveAllVars(NULL,"");
	bangcmddef *pBang;

	pBang = Bangs;
	while (pBang->Name != NULL) {
		RemoveBangCommand(pBang->Name);
		pBang++;
	}
    removeAllScriptBangs();
    varRemoveAll();

	SendMessage(hwndParent, LM_UNREGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);
	DestroyWindow(hwndMain);
	UnregisterClass(szAppName, dll);
}

//===========================================================================
// Module initialization, cleanup & !refresh / Support functions
//===========================================================================

int LoadSetup()
{
    char buffer[MAX_LINE_LENGTH];

    varSetFunc("xresolution",xRes);
    varSetFunc("yresolution",yRes);
	varSetFunc("bpp",bpp);

    varSetFunc("year",yearF);
    varSetFunc("month",monthF);
    varSetFunc("day",dayF);
    varSetFunc("weekday",weekdayF);
    varSetFunc("hour",hourF);
    varSetFunc("minute",minF);
    varSetFunc("second",secF);
	varSetFunc("milli",milliF);

    varSetFunc("mousex",mousexF);
    varSetFunc("mousey",mouseyF);

	bOldSyntax = GetRCBoolDef("mzScriptOldSyntax", FALSE);
	bNoStartOnRefresh = GetRCBoolDef("mzNoStartOnRefresh", FALSE);
	bReadStep = GetRCBoolDef("mzScriptUseStep", FALSE);
	bNoCreate = GetRCBoolDef("mzScriptNoCreate", FALSE);

	GetRCString("mzAutosaveVars",buffer,"",MAX_LINE_LENGTH);
	strtok(buffer," \t");
	if (!nstrcmp(buffer,"delayed")) autosave=1;
	else if (!nstrcmp(buffer,"immediately")) autosave=2;

	GetRCString("mzBangChar",buffer,"|",MAX_LINE_LENGTH);

    if (strlen(buffer)>0)
        bangChar = buffer[0];

    GetRCString("mzListSeparator",buffer,":",MAX_LINE_LENGTH);
    if (strlen(buffer)>0)
        listSep = buffer[0];

	FILE *f;
	f = LCOpen(NULL);		// everything but NULL breaks pLS support atm
	if(f)
	{
		char* tokens[2];
		char filename[MAX_LINE_LENGTH];
		char setting[MAX_LINE_LENGTH];

		tokens[0] = setting;
		tokens[1] = filename;
		filename[0] = 0;
		while(LCReadNextConfig(f, "*mzVarFile", buffer, sizeof(buffer)))
		{
			LCTokenize(buffer, tokens, 2, NULL);
			readVarFile(filename);
		}
		LCClose(f);
		f = LCOpen(NULL);	// file needs to be reopened... wtf ??????
		while(LCReadNextConfig(f, "*mzScriptFile", buffer, sizeof(buffer)))
		{
			LCTokenize(buffer, tokens, 2, NULL);
			readScriptFile(hwndParent, filename);
		}
		LCClose(f);
	} else return 2;
	return 0;
}

//===========================================================================
// Variable routines
//===========================================================================

// Does var exist?
int varExist(char *name)
{
    t_varlist *tmp=vars;
    while (tmp) {
        if (!nstrcmp(tmp->name,name))
            return 1;
        tmp=tmp->next;
    }
    return 0;
}

//--------------------

// Remove var from var list
void varRemove(char *name)
{
    t_varlist **tmp=&vars,*t2;

    while (*tmp) {
        if (!nstrcmp((*tmp)->name,name)) {
            free((*tmp)->name);
            if ((*tmp)->value)
                free((*tmp)->value);
            if ((*tmp)->varFile)
                free((*tmp)->varFile);
            t2 = *tmp;
            *tmp = (*tmp)->next;
            free(t2);
            return;
        }
        tmp=&((*tmp)->next);
    }
    return;
}

//--------------------

// Sets value to var
void varSet(char *name,char *val,char *filename = internalname)
{
	t_varlist *tmp=(t_varlist*)malloc(sizeof(t_varlist));

    if (!tmp) //Out of memory
        return;

	if (varExist(name)) {
		if (nstrcmp(internalname,getvarFile(name))) tmp->bUpdated = true;
		tmp->varFile = (char*)malloc(strlen(getvarFile(name))+1);
	    if (!tmp->varFile) {
			free(tmp);
	        return;
	    }
		strcpy(tmp->varFile,getvarFile(name));
	} else {
		tmp->bUpdated = false;
		tmp->varFile = (char*)malloc(strlen(filename)+1);
		if (!tmp->varFile) {
			free(tmp);
	        return;
	    }
	    strcpy(tmp->varFile,filename);
	}

    tmp->name = (char*)malloc(strlen(name)+1);
    if (!tmp->name) {
        free(tmp);
        return;
    }
    strcpy(tmp->name,name);
    if (val) {
        if ((tmp->value = strdup(val)) == NULL)
            return;
    } else
        tmp->value = NULL;

	tmp->func = NULL;

	varRemove(name);

	tmp->next = vars;
    vars = tmp;

	if (autosave==2 && tmp->bUpdated && nstrcmp(internalname,tmp->varFile))
		bangSaveVar(NULL,tmp->name);

    return;
}

//--------------------

// Set var to function
void varSetFunc(char *name,void (*f)(char *,int))
{
    t_varlist *tmp=(t_varlist*)malloc(sizeof(t_varlist));
    varRemove(name);

    if (!tmp) //Out of memory
        return;

    tmp->name = (char*)malloc(strlen(name)+1);
    if (!tmp->name) {
        free(tmp);
        return;
    }
    strcpy(tmp->name,name);
    tmp->varFile = (char*)malloc(strlen(internalname)+1);
    if (!tmp->varFile) {
        free(tmp);
        return;
    }
    strcpy(tmp->varFile,internalname);
    tmp->value = NULL;
    tmp->func = f;

    tmp->next = vars;
    vars = tmp;

    return;
}

//--------------------

// Get value from var
char *varGet(char *name)
{
    t_varlist *tmp=vars;
    while (tmp) {
        if (!nstrcmp(tmp->name,name)) {
            if (tmp->func) {
                tmp->func(returnBuffer,MAX_LINE_LENGTH);
                return returnBuffer;
            } else
                return tmp->value;
        }
        tmp=tmp->next;
    }
	return NULL;
}

char *getvarFile(char *name)
{
    t_varlist *tmp=vars;
    while (tmp) {
        if (!nstrcmp(tmp->name,name))
                return tmp->varFile;
        tmp=tmp->next;
    }
    return internalname;
}

//--------------------

// Clear all vars
void varRemoveAll(void)
{
    t_varlist *tmp=vars,*t2;
    vars = NULL;

    while (tmp) {
        t2=tmp->next;
        free(tmp->name);
        if (tmp->value)
            free(tmp->value);
        if (tmp->varFile)
            free(tmp->varFile);
        free(tmp);
        tmp=t2;
    }

    return;
}

//--------------------

// Expanding parser function
int varExpand(char *dest, const char *firstSrc, int n)
{
    char varName[256];
	char *tmp;
    char buffer[MAX_LINE_LENGTH];
	char *src;
	int i;
	int type;
	
	char ENCAP1;
	char ENCAP2;

    if (bOldSyntax == true)
	{
        ENCAP1 = '[';
        ENCAP2 = ']';
	}
    else
	{
        ENCAP1 = '{';
        ENCAP2 = '}';
	}

	src = buffer;

	VarExpansion(src, firstSrc);		// might be obsolete

    while (*src && n>1)
        if (*src=='%' && *(src+1)==ENCAP1) {
            i=0;
            src+=2;
            while (*src!=ENCAP2 && *src!=':' && *src && i<255)
                varName[i++] = *src++;
            varName[i]=0;

            if (*src) {
                type=0;
                if (*src==':') {
                    src++;
                    if (*src=='_') {
                        type=2;
                        src++;
                    } else
                        type=1;
                }
                if (*src==ENCAP2) src++;

                tmp=varGet(varName);
                if (tmp) {
                    if (type==1) {
                        while (*tmp && *tmp!=listSep && n>1) {
                            n--;
                            *dest++=*tmp++;
                        }
                    } else {
                        if (type==2) {
                            while (*tmp && *tmp!=listSep && n>1) tmp++;
                            if (*tmp==listSep) tmp++;
                        }
                        while (*tmp && n>1) {
                            n--;
                            *dest++=*tmp++;
                        }
                    }
                }
            }
        } else if (*src=='?' && *(src+1)==ENCAP1) {
            i=0;
            while (*src && *src!='\"') src++;
            if (!*src) continue;
            src++;
            tmp=src;
            while (*src && *src!='\"') src++;
            if (!*src) continue;
            *src++=0;
//            if (MessageBox(NULL, tmp, szAppName, MB_OKCANCEL | MB_ICONQUESTION | MB_APPLMODAL)==IDOK) {
            if (MessageBox(NULL, tmp, szAppName, MB_OKCANCEL | MB_ICONQUESTION | MB_TOPMOST | MB_SETFOREGROUND)==IDOK) {
                while (*src && *src!='\"') src++;
                if (!*src) continue;
                tmp=++src;
                while (*src && *src!='\"') src++;
                if (!*src) continue;
                *src++=0;
                while(*tmp && n>1) {
                    n--;
                    *dest++=*tmp++;
                }
            } else {
                i=0;
                while (*src && i<3) if (*src++=='\"') i++;
                if (!*src) continue;
                tmp=src;
                while (*src && *src!='\"') src++;
                if (!*src) continue;
                *src++=0;
                while(*tmp && n>1) {
                    n--;
                    *dest++=*tmp++;
                }
            }
            while (*src && *src!=ENCAP2) src++;
            if (*src) src++;
        } else {
            n--;
            *dest++=*src++;
        }

    if (n!=0) *dest=0;
	return 0;
}


//===========================================================================
// Script routines
//===========================================================================

void addBangExec(char *bang,char *exec,LINETYPE l)
{
    t_banglist **tmp=&bangs,*tmp2;
    t_execlist **exectmp;

    while(*tmp) {
        if (!nstrcmp((*tmp)->name,bang)) {
            exectmp = &((*tmp)->toexec);

            while(*exectmp)
                exectmp = &((*exectmp)->next);
            *exectmp = (t_execlist*)malloc(sizeof(t_execlist));
            if(!(*exectmp))
                return;
            (*exectmp)->next = NULL;
            (*exectmp)->line = l;
            if (exec) {
                (*exectmp)->exec = strdup(exec);
                if (!((*exectmp)->exec)) {
                    free(*exectmp);
                    *exectmp = NULL;
                    return;
                }
            } else
                (*exectmp)->exec = NULL;

            return;
        }
        tmp=&((*tmp)->next);
    }

    //Bang not present already
    tmp2 = (t_banglist*)malloc(sizeof(t_banglist));
    if (!tmp2)
        return;

    tmp2->next=bangs;
    tmp2->refcount=0;
    tmp2->todelete=0;
    tmp2->next=bangs;

    tmp2->name=strdup(bang);
    if (!(tmp2->name)) {
        free(tmp2);
        return;
    }
    tmp2->toexec = (t_execlist*)malloc(sizeof(t_execlist));
    if (!(tmp2->toexec)) {
        free(tmp2->name);
        free(tmp2);
        return;
    }
    tmp2->toexec->next=NULL;
    tmp2->toexec->line=l;
    if (exec) {
        tmp2->toexec->exec = strdup(exec);
        if (!(tmp2->toexec->exec)){
            free(tmp2->toexec);
            free(tmp2->name);
            free(tmp2);
            return;
        }
    } else
        tmp2->toexec->exec = NULL;
    bangs = tmp2;
}

//--------------------

int existScriptBang(char *name)
{
    t_banglist *bangtmp=bangs;
    while(bangtmp) {
        if (!nstrcmp(bangtmp->name,name)) return 1;
        bangtmp = bangtmp->next;
    }
    return 0;
}

//--------------------

void freeScriptBang(t_banglist *bangtmp)
{
    t_execlist *exectmp,*exectmp2;

    exectmp = bangtmp->toexec;

    while(exectmp) {
        exectmp2 = exectmp->next;
        if (exectmp->exec) free(exectmp->exec);
        free(exectmp);
        exectmp = exectmp2;
    }

    RemoveBangCommand(bangtmp->name);
    free(bangtmp->name);
    free(bangtmp);
}

//--------------------

void removeScriptBang(char *name)
{
    t_banglist **bangtmp=&bangs,*bangtmp2;

    while(*bangtmp) {
        if (!nstrcmp((*bangtmp)->name,name)) {
            bangtmp2=(*bangtmp);
            (*bangtmp)=(*bangtmp)->next;

            if (bangtmp2->refcount) {
                bangtmp2->todelete=1;
                return;
            }

            freeScriptBang(bangtmp2);

            return;
        }
        bangtmp = &((*bangtmp)->next);
    }
}

//--------------------

void removeAllScriptBangs(void)
{
    t_banglist *bangtmp=bangs,*bangtmp2;
    t_execlist *exectmp,*exectmp2;

    bangs = NULL;

    while(bangtmp) {
        bangtmp2=bangtmp->next;
        exectmp = bangtmp->toexec;
        while(exectmp) {
            exectmp2 = exectmp->next;
            if (exectmp->exec) free(exectmp->exec);
            free(exectmp);
            exectmp = exectmp2;
        }
        RemoveBangCommand(bangtmp->name);
        free(bangtmp->name);
        free(bangtmp);
        bangtmp = bangtmp2;
    }
}

//===========================================================================
// Internal functions (to return things like screen width)
//===========================================================================

void xRes(char *place,int max) {
    sprintf(place,"%d",GetSystemMetrics(SM_CXSCREEN));
}

void yRes(char *place,int max) {
    sprintf(place,"%d",GetSystemMetrics(SM_CYSCREEN));
}

void bpp(char *place,int max) {

	HKEY hkey;
	DWORD dwType, dwSize;
	char bitspp[3];

	strcpy(bitspp, "16");	// default value in case opening the regkey doesn't work

	if(RegOpenKeyEx(HKEY_CURRENT_CONFIG, "Display\\Settings", NULL, KEY_READ, &hkey) == ERROR_SUCCESS)
	{
		dwType = REG_SZ;
		dwSize = sizeof(bitspp);
		RegQueryValueEx(hkey, "BitsPerPixel", NULL, &dwType, (PBYTE)&bitspp, &dwSize);
		RegCloseKey(hkey);
	}

	sprintf(place,"%s",bitspp);
}

void yearF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wYear);
}

void monthF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wMonth);
}

void dayF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wDay);
}

void weekdayF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wDayOfWeek);
}

void hourF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wHour);
}


void minF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wMinute);
}

void secF(char *place,int max) {
    SYSTEMTIME systemTime;
    GetLocalTime(&systemTime);
    sprintf(place,"%d",systemTime.wSecond);
}

//Added by Brian Hartvigsen
void milliF(char *place,int max) {
	SYSTEMTIME systemTime;
	GetLocalTime(&systemTime);
	sprintf(place,"%d",systemTime.wMilliseconds);
}

void mousexF(char *place,int max) {
    POINT p;
    GetCursorPos(&p);
    sprintf(place,"%d",p.x);
}

void mouseyF(char *place,int max) {
    POINT p;
    GetCursorPos(&p);
    sprintf(place,"%d",p.y);
}

//===========================================================================

void ExecCommand(HWND caller,char *szSample)
{
    char *newcmd, *args, *p;
    char command[MAX_LINE_LENGTH];
    char dir[MAX_LINE_LENGTH], full_dir[MAX_LINE_LENGTH];
    SHELLEXECUTEINFO si;

    if(!szSample || strlen(szSample)<1)
        return;

    strcpy(command,szSample);

    if(*command == '\"') {
        newcmd = command + sizeof(char);
        newcmd = strtok(newcmd, "\"");
        args = strtok(NULL, "\0");

        if(!newcmd) {
            errorMsg("Missing closing quote.");
            return;
        }
    } else {
        newcmd = strtok(command, "\t ");
        if(!newcmd) {
            newcmd = command;
            args = NULL;
        } else {
            args = strtok(NULL, "\0");
        }
    }

    if (*newcmd==bangChar) *newcmd='!';

    p = newcmd + (strlen(newcmd) - 1)*sizeof(char);
    while(*newcmd == ' ' || *newcmd == '\t')
        ++newcmd;
    while(*p == ' ' || *p == '\t') {
        *p = '\0';
        --p;
    }

    if(*newcmd != '!') {
        if(args) {
            p = args + (strlen(args) - 1)*sizeof(char);
            while(*args == ' ' || *args == '\t' || *args == '\"')
                ++args;
            while(*p == ' ' || *p == '\t' || *p == '\"') {
                *p = '\0';
                --p;
            }
        }
    } else {
        if(args) {
            p = args + (strlen(args) - 1)*sizeof(char);
            while(*args == ' ' || *args == '\t')
                ++args;
            while(*p == ' ' || *p == '\t') {
                *p = '\0';
                --p;
            }
        }
    }

    if(*newcmd == '!') {
        ParseBangCommand(caller, newcmd, args);
    } else {
        memset(&si, 0, sizeof(si));
        si.cbSize = sizeof(SHELLEXECUTEINFO);
        si.hwnd = caller;
        si.lpFile = newcmd;
        si.lpParameters = args;
        si.nShow = SW_SHOWNORMAL;
        si.fMask = SEE_MASK_DOENVSUBST;
        _splitpath(newcmd, full_dir, dir, NULL, NULL);
        si.lpVerb = NULL;
        si.lpDirectory = full_dir;
        strcat(full_dir, dir);
        ShellExecuteEx(&si);
    }
}

//===========================================================================
// The bangs...
//===========================================================================

void execFirst(HWND caller,char *expr)
{
    int count;
    char *tmp;
    if (!expr) return;
    if (*expr=='\'') {
        expr++;
        expr = strtok(expr, "\'");
        if(!expr) {
            errorMsg("Missing closing \'.");
            return;
        }
        ExecCommand(caller,expr);
    } else if (*expr=='{') {
        expr++;
        tmp=expr;
        count=1;
        while (count) {
            if (*tmp=='{') count++;
            if (*tmp=='}') count--;
            if (!*tmp) {
                errorMsg("Missing closing }.");
                return;
            }
            if (count) tmp++;
        }
        *tmp=0;

        ExecCommand(caller,expr);
    } else
        ExecCommand(caller,expr);
}

//--------------------

void execSecond(HWND caller,char *expr)
{
    char *tmp;
    int count;
    if (!expr) return;
    if (*expr=='\'') {
        expr++;
        expr = strtok(expr, "\'");
        if(!expr) {
            errorMsg("Missing closing quote \'.");
            return;
        }
        expr = strtok(NULL, "\'");
        if (!expr) return;
        expr = strtok(NULL, "\'");
        if(!expr) {
            errorMsg("Missing else closing quote \'.");
            return;
        }
        ExecCommand(caller,expr);
    } else if (*expr=='{') {
        expr++;

        tmp=expr;
        count=1;
        while (count) {
            if (*tmp=='{') count++;
            if (*tmp=='}') count--;
            if (!*tmp) {
                errorMsg("Missing closing }.");
                return;
            }
            tmp++;
        }
        while (*tmp && *tmp!='{') tmp++;
        if (!*tmp) return;

        expr=++tmp;
        count=1;
        while (count) {
            if (*tmp=='{') count++;
            if (*tmp=='}') count--;
            if (!*tmp) {
                errorMsg("Missing closing }.");
                return;
            }
            if (count) tmp++;
        }

        *tmp=0;
        ExecCommand(caller,expr);
    }
}

//--------------------

void bangSetVar(HWND caller, const char *args)
{
    char newvar[MAX_LINE_LENGTH], val[MAX_LINE_LENGTH];
	char command[MAX_LINE_LENGTH];
	char* tokens[2];
	int count;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args, MAX_LINE_LENGTH);

	newvar[0] = val[0] = 0;
	tokens[0] = newvar;
	tokens[1] = val;
	
	count = LCTokenize(command, tokens, 2, NULL);

	if (count <= 1) {
		errorMsg("Parameter missing");
		return;
	}
	varSet(newvar,val);

	return;
}

//--------------------
void bangAddVar(HWND caller, const char *args)
{
    char *newvar, *val, *tmp, command[MAX_LINE_LENGTH];
    int i,j,sign;
    double f,d;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    newvar = strtok(command, "\t ");

    if(!newvar) return;

    if ((val = strtok(NULL, "\0")) == NULL) return;
	while (*val==' ' || *val=='\t') val++;
    if (*val=='\"') {
        val++;
        if (*val!='\"')
            val = strtok(val, "\"");
        else
            *val=0;
        if (!val) {
            errorMsg("Missing closing quote.");
            return;
        }
    }

    tmp = varGet(newvar);
    if (!tmp)
		if(bNoCreate)
			return;
		else {
			char *temp = "0";
			varSet(newvar, temp);
			tmp = varGet(newvar);
		}
	

// Rounding up check.

    i = atoi(tmp);
    f = atof(tmp);
    d = f-i;
    if(d >= 0.5)
        i=i+1;

    j = atoi(val);
    f = atof(val);
    d = f-j;
    if(d >= 0.5)
        j=j+1;

//    

    i = i+j;
    sign = i<0;
    if (sign) i=-i;

    tmp=command+MAX_LINE_LENGTH-1;
    *tmp=0;
    if (!i) *--tmp='0';
    while (i && tmp>command) {
        *--tmp='0'+(i%10);
        i/=10;
    }
    if (sign) *--tmp='-';

    varSet(newvar,tmp);

    return;
}

//--------------------

void bangMulVar(HWND caller, const char *args)
{
    char *newvar, *val, *tmp, command[MAX_LINE_LENGTH];
    int i,sign;
    float f;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    newvar = strtok(command, "\t ");

    if(!newvar)
		return;

	if ((val = strtok(NULL, "\0")) == NULL) return;
	while (*val==' ' || *val=='\t') val++;
    if (*val=='\"') {
        val++;
        if (*val!='\"')
            val = strtok(val, "\"");
        else
            *val=0;
        if (!val) {
            errorMsg("Missing closing quote.");
            return;
        }
    }

    tmp = varGet(newvar);
    if (!tmp)
		if(bNoCreate)
			return;
		else {
			char *temp = "0";
			varSet(newvar, temp);
			tmp = varGet(newvar);
		}

    sscanf(val,"%f",&f);

    i = (int)((atoi(tmp))*f);
    sign = i<0;
    if (sign) i=-i;

    tmp=command+MAX_LINE_LENGTH-1;
    *tmp=0;
    if (!i) *--tmp='0';
    while (i && tmp>command) {
        *--tmp='0'+(i%10);
        i/=10;
    }
    if (sign) *--tmp='-';

    varSet(newvar,tmp);

    return;
}

//Added by Brian "Tres`ni" Hartvigsen
void bangModVar(HWND caller, const char *args)
{
    char *newvar, *val, *tmp, command[MAX_LINE_LENGTH];
    int i/*,sign;
    float f*/;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    newvar = strtok(command, "\t ");

    if(!newvar) return;

    if ((val = strtok(NULL, "\0")) == NULL) return;
	while (*val==' ' || *val=='\t') val++;
    if (*val=='\"') {
        val++;
        if (*val!='\"')
            val = strtok(val, "\"");
        else
            *val=0;
        if (!val) {
            errorMsg("Missing closing quote.");
            return;
        }
    }

    tmp = varGet(newvar);
    if (!tmp)
        if(bNoCreate)
			return;
		else {
			char *temp = "0";
			varSet(newvar, temp);
			tmp = varGet(newvar);
		}

    //sscanf(val,"%f",&f);


	if (atoi(val) != 0) 
		return;
	
	i = (int)((atoi(tmp))%atoi(val));

    /*sign = i<0;*/
    if (/*sign*/i<0) i=-i;
    tmp=command+MAX_LINE_LENGTH-1;
    *tmp=0;
    if (!i) *--tmp='0';
    while (i && tmp>command) {
        *--tmp='0'+(i%10);
        i/=10;
    }
    //if (sign) *--tmp='-';

    varSet(newvar,tmp);

    return;
}

//Added By Brian "Tres`ni" Hartvigsen
void bangRndVar(HWND caller, const char *args)
{
    char *newvar, *val, *tmp, command[MAX_LINE_LENGTH];
    int i;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    newvar = strtok(command, "\t ");

    if(!newvar) return;

    if ((val = strtok(NULL, "\0")) == NULL) return;
	while (*val==' ' || *val=='\t') val++;
    if (*val=='\"') {
        val++;
        if (*val!='\"')
            val = strtok(val, "\"");
        else
            *val=0;
        if (!val) {
            errorMsg("Missing closing quote.");
            return;
        }
    }

    tmp = varGet(newvar);
    if (!tmp)
        if(bNoCreate)
			return;
		else {
			char *temp = "0";
			varSet(newvar, temp);
			tmp = varGet(newvar);
		}

    //sscanf(val,"%f",&f);

	if (atoi(val) == 0)
		return;

    i = 1 + rand() % atoi(val);

    if (i<0) i=-i;
    tmp=command+MAX_LINE_LENGTH-1;
    *tmp=0;
    if (!i) *--tmp='0';
    while (i && tmp>command) {
        *--tmp='0'+(i%10);
        i/=10;
    }

    varSet(newvar,tmp);

    return;
}

//--------------------

void bangIntVar(HWND caller, const char *args)
{
    char *newvar, *tmp, command[MAX_LINE_LENGTH];
    double f,d;
    int i,sign;

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    newvar = strtok(command, "\t ");

    if(!newvar) return;

    tmp = varGet(newvar);
    if (!tmp)
        return;

// Rounding up check.

    i = atoi(tmp);
    f = atof(tmp);
    d = f-i;
    if(d >= 0.5)
        i=i+1;

//    

//  i = atoi(tmp);
    sign = i<0;
    if (sign) i=-i;

    tmp=command+MAX_LINE_LENGTH-1;
    *tmp=0;
    if (!i) *--tmp='0';
    while (i && tmp>command) {
        *--tmp='0'+(i%10);
        i/=10;
    }
    if (sign) *--tmp='-';

    varSet(newvar,tmp);
}

//--------------------

void bangShowVar(HWND caller, const char *args)
{
    char *var, command[MAX_LINE_LENGTH], val[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if(!var) {
        var = command;
    }
    if (varExist(var)) {
        sprintf(val,"var \"%s\" exists.\nvalue = %s\nfile = %s",var,varGet(var),getvarFile(var));
    } else {
        sprintf(val,"var \"%s\" does not exist.",var);
    }
//    MessageBox(NULL, val, szAppName, MB_OK | MB_APPLMODAL);
    MessageBox(NULL, val, szAppName, MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    return;
}

//--------------------

void bangMsgBox(HWND caller, const char *args)
{

    char command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
//    MessageBox(NULL, command, szAppName, MB_OK | MB_TOPMOST | MB_APPLMODAL);
    MessageBox(NULL, command, szAppName, MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
//  MessageBox(hCaller, szMessage, szTitle, MB_OK | MB_TOPMOST | MB_SETFOREGROUND);
    return;
}

//--------------------

void bangRemoveVar(HWND caller, const char *args)
{
    char *var, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if(!var) {
        var = command;
    }
    varRemove(var);
    return;
}

//--------------------

void bangRunVar(HWND caller, const char *args)
{
    char *var, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if(!var) {
        var = command;
    }
    ExecCommand(caller,varGet(var));

    return;

}

//--------------------

void bangSaveVar(HWND caller, const char *args)
{
	char buffer[MAX_LINE_LENGTH];

	varExpand(buffer, args, MAX_LINE_LENGTH);
	if (varExist(buffer)) {
		ParseFile vF(getvarFile(buffer));
		while (nstrcmp(vF.NextToken(), buffer)) {
			vF.SkipLine();
		}
		vF.NextToken();
		vF.SetToken(varGet(buffer));
		vF.Put();
	}
}

//--------------------

void bangSaveAllVars(HWND caller, const char *args)
{
	t_varlist *tmp=vars;

	while(tmp) {
		if (tmp->bUpdated && nstrcmp(internalname, tmp->varFile)) {
			ParseFile vF(tmp->varFile);
			while (nstrcmp(vF.NextToken(), tmp->name)) {
				vF.SkipLine();
			}
			vF.NextToken();
			vF.SetToken(tmp->value);
			vF.Put();
		}
		tmp=tmp->next;
	}
}

//--------------------

void bangVarDump(HWND caller, const char *args)
{
    t_varlist *tmp=vars;
	char buffer[MAX_LINE_LENGTH];

    varExpand(buffer, args, MAX_LINE_LENGTH);	
	
	FILE* file = fopen(buffer, "a");

	while(tmp) {					// the code below is crap but it works
		strcpy(buffer, tmp->name);
		strcat(buffer, "\t");
		if (strlen(tmp->name) <= 7) strcat(buffer, "\t");
		strcat(buffer, varGet(tmp->name));
		strcat(buffer, "\t");
		if (strlen(varGet(tmp->name)) <= 7) strcat(buffer, "\t");
		strcat(buffer, tmp->varFile);
		strcat(buffer, "\n");
		fputs(buffer,file);
		tmp=tmp->next;
	}
	
	fclose(file);
	
	return;
}

//--------------------

void bangIfExist(HWND caller, const char *args)
{

    char *var, *val, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if(!var) {
        var = command;
        val = NULL;
    } else {
        val = strtok(NULL, "\0");
    }

    if (varExist(var))
        execFirst(caller,val);
    else
        execSecond(caller,val);

    return;
}

//--------------------

void bangIfNExist(HWND caller, const char *args)
{
    char *var, *val, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if(!var) {
        var = command;
        val = NULL;
    } else {
        val = strtok(NULL, "\0");
    }

    if (!varExist(var))
        execFirst(caller,val);
    else
        execSecond(caller,val);

    return;
}

//--------------------

void bangIfEq(HWND caller, const char *args)
{
    char *var, *comp, *exec, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if (var) {
        if ((comp = strtok(NULL, "\"")) == NULL) {
            errorMsg("Missing quote");
            return;
        }
        if ((exec = strtok(NULL, "\"")) == NULL) {
            errorMsg("Missing closing quote");
            return;
        }
        while ((*exec == ' ') || (*exec == '\t')) exec++;

        if (!nstrcmp(comp,varGet(var)))
            execFirst(caller,exec);
        else
            execSecond(caller,exec);
    }
    return;
}

//--------------------

void bangIfNEq(HWND caller, const char *args)
{

    char *var, *comp, *exec, command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);

    var = strtok(command, "\t ");
    if (var) {
        if ((comp = strtok(NULL, "\"")) == NULL) {
            errorMsg("Missing quote");
            return;
        }
        if ((exec = strtok(NULL, "\"")) == NULL) {
            errorMsg("Missing closing quote");
            return;
        }
        while ((*exec == ' ') || (*exec == '\t')) exec++;

        if (nstrcmp(comp,varGet(var)))
            execFirst(caller,exec);
        else
            execSecond(caller,exec);
    }
    return;
}

//--------------------

int evalIf(char *e,char **theRet)
{
    char *first,*second,*cmp,*tmp,*tmp2,*ret;

    if (!e) return 0;
    *theRet=e;

    tmp = e;
    while (*tmp && *tmp!='\"') tmp++;
    if (!*tmp) {
        errorMsg("Missing quote.");
        return 0;
    }
    first=++tmp;
    while (*tmp && *tmp!='\"') tmp++;
    if (!*tmp) {
        errorMsg("Missing quote.");
        return 0;
    }
    *tmp++=0;
    cmp=tmp;
    while (*tmp && *tmp!='\"') tmp++;
    if (!*tmp) {
        errorMsg("Missing quote.");
        return 0;
    }
    *tmp++=0;
    while (*cmp==' ' || *cmp=='\t') cmp++;
    tmp2 = cmp;
    while (*tmp2!=' ' && *tmp2!='\t') tmp2++;
    *tmp2=0;

    second = tmp;
    while (*tmp && *tmp!='\"') tmp++;
    if (!*tmp) {
        errorMsg("Missing quote.");
        return 0;
    }
    *tmp++=0;
    ret=tmp;
    while (*ret==' ' || *ret=='\t'  || *ret==')') ret++;
    *theRet = ret;
    if (!nstrcmp(cmp,"=")) {
		if (isnum(first) && isnum(second)) {
			if (atoi(first) == atoi(second))
				return 1;
		}
		else {
			if (!nstrcmp(first,second))
				return 1;
		}
    } else if (!nstrcmp(cmp,"<=")) {
		if (isnum(first) && isnum(second)) {
			if (atoi(first) <= atoi(second))
				return 1;
		}
		else {
			if (nstrcmp(first,second)<=0)
				return 1;
		}
    } else if (!nstrcmp(cmp,">=")) {
		if (isnum(first) && isnum(second)) {
			if (atoi(first) >= atoi(second))
				return 1;
		}
		else {
			if (nstrcmp(first,second)>=0)
				return 1;
		}
    } else if (!nstrcmp(cmp,"<")) {
        if (isnum(first) && isnum(second)) {
			if (atoi(first) < atoi(second))
				return 1;
		}
		else {
			if (nstrcmp(first,second)<0)
				return 1;
		}
    } else if (!nstrcmp(cmp,">")) {
        if (isnum(first) && isnum(second)) {
			if (atoi(first) > atoi(second))
				return 1;
		}
		else {
			if (nstrcmp(first,second)>0)
				return 1;
		}
    } else if (!nstrcmp(cmp,"<>")) {
        if (isnum(first) && isnum(second)) {
			if (atoi(first) != atoi(second))
				return 1;
		}
		else {
			if (nstrcmp(first,second))
				return 1;
		}
    }

    return 0;
}

//--------------------

void bangIfEval(HWND caller, const char *args)
{

    char *command, buffer[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(buffer,args,MAX_LINE_LENGTH);

    if (evalIf(buffer,&command))
        execFirst(caller,command);
    else
        execSecond(caller,command);

    return;
}

//--------------------

void bangScript(HWND caller, const char *bangname, const char *args)
{
	t_banglist *bangtmp=bangs;
	t_execlist *exectmp,*exectmp2;
	char *label,buffer[MAX_LINE_LENGTH],buffer2[MAX_LINE_LENGTH];

	varExpand(buffer,args,MAX_LINE_LENGTH);

	char *tmp2 = buffer2;
	char *tmp = buffer;

    while (*tmp) {				// escape code : %=
        if (*tmp=='%' && *(tmp+1)=='=') {
			*tmp2 = '\"';
			*tmp++;
		} else *tmp2 = *tmp;
		*tmp2++;
		*tmp++;
	}
	*tmp2 = 0;

	if (oldTokens > 0) {		// remove all old param vars
		varRemove("args");
		for (unsigned int i = 1; i<=oldTokens; i++) {
			sprintf(buffer,"\\%d",i);
			varRemove(buffer);
		}
		oldTokens = 0;	
	}
	
	if (nstrcmp(buffer2,"")) {	// set param vars if params given
		varSet("args",buffer2);

		char param[MAX_LINE_LENGTH], rest[MAX_LINE_LENGTH];

		char* token[1];

		token[0] = param;
		param[0] = 0;

		oldTokens++;

		LCTokenize(buffer2, token, 1, rest);

		varSet("\\1",param);
		
		while (nstrcmp(rest,"")) {
			strcpy(buffer2,rest);
			oldTokens++;
			sprintf(buffer,"\\%d",oldTokens);
			LCTokenize(buffer2, token, 1, rest);
			varSet(buffer,param);
		}
	}

	while(bangtmp) {			// original mzscript 0.8 code :)
		if (!nstrcmp(bangtmp->name,(char *)bangname)) {
            (bangtmp->refcount)++;
            exectmp = bangtmp->toexec;
            while(exectmp && exectmp->line!=EXIT) {
                label = buffer;
                varExpand(buffer,exectmp->exec,MAX_LINE_LENGTH);
                switch (exectmp->line) {
                    case EXEC:
                        ExecCommand(caller,buffer);
                        break;
                    case GOTOIF:
                        if (!evalIf(buffer,&label))
                            break;
                    case GOTO:
                        exectmp2 = bangtmp->toexec;
                        while (exectmp2 && !(exectmp2->line==LABEL &&
                                    !nstrcmp(exectmp2->exec,label)))
                            exectmp2 = exectmp2->next;
                        exectmp = exectmp2;
                        break;

                }
                if (exectmp) exectmp = exectmp->next;
            }
            (bangtmp->refcount)--;
            if ((!(bangtmp->refcount)) && (bangtmp->todelete))
                freeScriptBang(bangtmp);

            return;
        }
        bangtmp = bangtmp->next;
    }
}

//--------------------

//Execute command
void bangExec(HWND caller, const char *args)
{

    char command[MAX_LINE_LENGTH];

    if(!args || strlen(args)<1)
        return;

    varExpand(command, args,MAX_LINE_LENGTH);
    ExecCommand(caller,command);
    return;
}

//--------------------

// Remove bang script
void bangRemoveScript(HWND parent, const char *scriptname)
{
    char buffer[MAX_LINE_LENGTH];
    varExpand(buffer+1, scriptname,MAX_LINE_LENGTH-1);
    *buffer='!';
    removeScriptBang(buffer);
}

//--------------------

// Read script definition file
void readScriptFile(HWND parent, const char *scriptFile)
{
	if (strcmp(scriptFile,"")) {
		char buffer[MAX_LINE_LENGTH];
		varExpand(buffer, scriptFile, MAX_LINE_LENGTH);

		ParseFile sF(buffer);

		char rest[MAX_LINE_LENGTH];
		char currentBang[MAX_LINE_LENGTH];
		int i;

		currentBang[0] = 0;

		if (sF.isOk()) {
			while (strcmp(sF.NextToken(),"")) {
				if (!nstrcmp("*Script",sF.GetToken())) {
					strcpy(buffer, sF.NextToken());

					if (!nstrcmp("bang",buffer)) {
			            if (currentBang[0] == 0) {
			                strcpy(currentBang,sF.NextToken());
				            if (existScriptBang(currentBang))
					            removeScriptBang(currentBang);
						}
					} else {
						sF.GetLine();
						strcpy(rest, sF.GetToken());
						for(i=0;i<5;i++) {
							if ((!nstrcmp(buffer,par[i].command)) && (currentBang[0] != 0))
								addBangExec(currentBang,rest,par[i].type);
						}
						if ((!nstrcmp(buffer,"~bang")) && (currentBang[0] != 0)) {
							AddBangCommandEx(currentBang,bangScript);
					        currentBang[0] = 0;
					    }
//						if (!nstrcmp(buffer,"start"))
//					        ExecCommand(parent,rest);
					}
				} else {
					if (!nstrcmp("*mzScriptFile",sF.GetToken()))
						readScriptFile(parent, sF.NextToken());
					else if (!nstrcmp("*mzVarFile",sF.GetToken()))
						readVarFile(sF.NextToken());
				}

			}

			if (!(bRefresh && bNoStartOnRefresh)) {
				sF.Reset();
				while (strcmp(sF.NextToken(),""))
					if (!nstrcmp("*Script",sF.GetToken())) {
						strcpy(buffer, sF.NextToken());
						sF.GetLine();
						strcpy(rest, sF.GetToken());
						if (!nstrcmp(buffer,"start"))
						    ExecCommand(parent,rest);
					}
			}
		} else {
			sprintf(buffer,"ScriptFile \"%s\" not found",scriptFile);
			errorMsg(buffer);
		}
	} else errorMsg("ScriptFile \"\" not found");
}

// legacy step.rc support
void readStep(HWND parent)
{
    char buffer[MAX_LINE_LENGTH];
    char *tmp,*var, *exec, command[MAX_LINE_LENGTH],currentBang[MAX_LINE_LENGTH];
    FILE *f;
    int i;

    f = LCOpen (NULL);

    if (f) {
        buffer[0] = 0;
        currentBang[0] = 0;
        while (LCReadNextConfig (f, "*Script", buffer, sizeof (buffer)))
        {

            if(strlen(buffer)<1)
                continue;

            tmp=buffer;
            while (tmp>=buffer && (*tmp==' ' || *tmp=='\t')) *tmp--=0;

            VarExpansion(command, buffer);

            if ((var = strtok(command, "\t ")) == NULL)
                continue;

            var = strtok(NULL, "\t ");
            if(!var) {
                var = command;
                exec = NULL;
            } else {
                exec = strtok(NULL, "\0");
            }
            if (exec) while (*exec==' ' || *exec=='\t') exec++;

            if (!nstrcmp(var,"start"))
                ExecCommand(parent,exec);

            if (!nstrcmp(var,"var"))
                bangSetVar(parent,exec);

            if ((!nstrcmp(var,"bang")) && (currentBang[0] == 0)) {
                strcpy(currentBang,exec);
                if (existScriptBang(currentBang))
                    removeScriptBang(currentBang);
            }

            for(i=0;i<5;i++) {
                if ((!nstrcmp(var,par[i].command)) && (currentBang[0] != 0))
                    addBangExec(currentBang,exec,par[i].type);
            }

            if ((!nstrcmp(var,"~bang")) && (currentBang[0] != 0)) {
				AddBangCommandEx(currentBang,bangScript);
                currentBang[0] = 0;
            }
		}
		LCClose(f);
	} else
		errorMsg("No file");

}

// Read var definition file
void readVarFile(char *varFile)
{
	char buffer[MAX_LINE_LENGTH];

	ParseFile vF(varFile);
	if (vF.isOk())
		while (strcmp(vF.NextToken(),"")) {
			strcpy(buffer, vF.GetToken());
			varSet(buffer, vF.NextToken(), varFile);
			vF.SkipLine();
		}
	else {
		sprintf(buffer,"VarFile \"%s\" not found",varFile);
		errorMsg(buffer);
	}
}

//--------------------

void bangSetListSep(HWND parent, const char *arg)
{
	if (arg)
		if (*arg) listSep = *arg;

}

//--------------------

void pause(HWND parent, LPCTSTR pszArgs )
{
	unsigned long lInterval = atoi( pszArgs );
	unsigned long lTime = GetTickCount();
	MSG msg;
	
	if( lInterval > 0 )
	{
		while( (GetTickCount() - lTime) < lInterval )
		{
			if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
	}
}


//===========================================================================
// Message handling
//===========================================================================

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{

		case LM_GETREVID:
		{
			char *buf = (char *) lParam;

			if (wParam == 0)
			{
				strcpy(buf, szVersion);
			}
			else if (wParam == 1)
			{
				strcpy(buf, szDetailed);
			} else
			{
				strcpy(buf, "");
			}
			return strlen(buf);
		}
		break;

		case LM_REFRESH:
		{
			bRefresh = true;
			if (autosave==1) bangSaveAllVars(NULL,"");
		    if (bReadStep) readStep(hwnd);
			removeAllScriptBangs();
			varRemoveAll();
			LoadSetup();
			bRefresh = false;
		}
		break;

	}
	return DefWindowProc(hwnd,message,wParam,lParam);
}

//===========================================================================
